MANIFEST_BUILD: str = "1.21.111"  # The build version of the manifest.
BLOCK_SERVER_VERSION: str = MANIFEST_BUILD  # The version of the block server.
ENTITY_SERVER_VERSION: str = MANIFEST_BUILD  # The version of the entity server.
ENTITY_CLIENT_VERSION: str = MANIFEST_BUILD  # The version of the entity client.
ITEM_SERVER_VERSION: str = MANIFEST_BUILD  # The version of the item server.
RECIPE_JSON_FORMAT_VERSION = MANIFEST_BUILD  # The version of the recipe json format.
SPAWN_RULES_VERSION: str = MANIFEST_BUILD  # The version of the spawn rules.
SOUND_DEFINITIONS_VERSION: str = MANIFEST_BUILD  # The version of the sound definitions.
FOG_VERSION: str = MANIFEST_BUILD  # The version of the fog.
JIGSAW_VERSION: str = MANIFEST_BUILD  # The version of the jigsaw.
BIOME_SERVER_VERSION: str = MANIFEST_BUILD  # The version of the biome server.
BIOME_CLIENT_VERSION: str = MANIFEST_BUILD  # The version of the biome client.


ANIMATION_CONTROLLERS_VERSION: str = "1.10.0"  # The version of the animation controllers.
BP_ANIMATION_VERSION: str = "1.10.0"  # The version of the behavior pack animation.
RP_ANIMATION_VERSION: str = "1.10.0"  # The version of the resource pack animation.
RENDER_CONTROLLER_VERSION: str = "1.10.0"  # The version of the render controller.

GEOMETRY_VERSION: str = "1.21.20"  # The version of the geometry.
DIALOGUE_VERSION: str = "1.18.0"  # The version of the dialogue.
MATERIALS_VERSION: str = "1.0.0"  # The version of the materials.
CAMERA_PRESET_VERSION = "1.21.0"  # The version of the camera presets.
BLOCK_JSON_FORMAT_VERSION = "1.1.0"  # The version of the block json format.
CRAFTING_ITEMS_CATALOG = "1.21.60"  # The version of the crafting items catalog.
BLOCK_CULLING_VERSION = "1.20.80"  # The version of the block culling rules.

PBR_SETTINGS_VERSION: str = MANIFEST_BUILD  # The version of the atmospheric effects.

MODULE_MINECRAFT_SERVER: str = "2.2.0"  # The version of the Minecraft server module.
MODULE_MINECRAFT_SERVER_UI: str = "2.0.0"  # The version of the Minecraft UI module.
MODULE_MINECRAFT_SERVER_EDITOR: str = "0.1.0"  # The version of the Minecraft UI module.
MODULE_MINECRAFT_SERVER_GAMETEST: str = "1.0.0"  # The version of the Minecraft UI module.
